from openapi_typed_2 import Schema, Reference
from dataclasses import dataclass, field
from typing import Sequence, Union
import enum

####################################
#### does a diff on two schemas

class SM(enum.Enum):
    SEQUENCE = 0
    MAPPING = 1


@dataclass(frozen=True)
class SchemaDiff:
    differing_types: Sequence[Sequence[Union[str, SM]]] = field(default_factory=list)
    differing_keys: Sequence[Sequence[Union[str, SM]]] = field(default_factory=list)

def make_schema_diff(s0: Union[Schema, Reference], s1: Union[Schema, Reference], path: Sequence[Union[str, SM]] = []) -> SchemaDiff:
    if isinstance(s0, Schema) and isinstance(s1, Schema):
        if s0._type != s1._type:
            return SchemaDiff(differing_types=[path])
        elif s0._type == 'object':
            props = [set(x.keys()) if x is not None else set([]) for x in [s0.properties, s1.properties]]
            kdiff = props[0].symmetric_difference(props[1])
            diffs = [make_schema_diff(s0.properties[p], s1.properties[p]) for p in props[0].intersection(props[1])]
            return SchemaDiff(differing_types=sum([x.differing_types for x in diffs], []), differing_keys=sum([x.differing_keys for x in diffs], [[x, *path] for x in kdiff]))
        elif s0._type == 'array':
            if type(s0.items) != type(s1.items):
                return SchemaDiff(differing_types=[[SM.SEQUENCE, *path]])
            elif isinstance(s0.items, Schema) and isinstance(s1.items, Schema):
                return make_schema_diff(s0.items, s1.items, path=[SM.SEQUENCE, *path])
            else:
                # a tuple or some weird state we didn't predit
                # for now we give up as it is so rare,
                # but should inspect this in the future
                return SchemaDiff()
        return SchemaDiff()
    elif isinstance(s0, Reference) and isinstance(s1, Reference):
        return SchemaDiff(differing_types=[path] if s0._ref != s1._ref else [])
    else:
        return SchemaDiff(differing_types=[path])


_SCHEMA_DEFAULT = {'title': None, 'multipleOf': None, 'maximum': None, 'exclusiveMaximum': None, 'minimum': None, 'exclusiveMinimum': None, 'maxLength': None, 'minLength': None, 'pattern': None, 'maxItems': None, 'minItems': None, 'uniqueItems': None, 'maxProperties': None, 'minProperties': None, 'required': None, 'enum': None, 'allOf': None, 'oneOf': None, 'anyOf': None, 'items': None, 'properties': None, 'additionalProperties': None, 'description': None, 'default': None, 'nullable': None, 'discriminator': None, 'readOnly': None, 'writeOnly': None, 'example': None, 'externalDocs': None, 'deprecated': None, 'xml': None, '_format': None, '_type': None, '_not': None, '_x': None}


from functools import reduce
import lenses
from lenses.optics.base import Prism
lens = lenses.lens
import jsonschema
from ..schemabuilder.operation import operation_from_string
from dataclasses import replace
import json
import re
from http_types import Request
from typing import cast, Sequence, Tuple, TypeVar, Callable, Optional, Mapping, Union, Any
from openapi_typed_2 import convert_from_openapi, OpenAPIObject, Responses, MediaType, Response, RequestBody, Header, Operation, Parameter, Components, Reference, Schema, PathItem
from urllib.parse import urlparse
from ..schemabuilder.defaults import _SCHEMA_DEFAULT

C = TypeVar('C') 
D = TypeVar('D')
S = TypeVar('S')
T = TypeVar('T')
W = TypeVar('W')
X = TypeVar('X')
Y = TypeVar('Y')
Z = TypeVar('Z')

__all__ = ['matcher', 'faker', 'change_ref', 'change_refs']


#def pp(s, c: C) -> C:
#    print(s, c)
#    return c

all_methods: Sequence[str] = [
  "get",
  "post",
  "put",
  "delete",
  "options",
  "head",
  "patch",
  "trace",
]

def omit_method_from_path_item(p: PathItem, a: str) -> PathItem:
    return replace(
        p, get=None
    ) if a == "get" else replace(
        p, post=None
    ) if a == "post" else replace(
        p, put=None
    ) if a == "put" else replace(
        p, delete=None
    ) if a == "delete" else replace(
        p, patch=None
    ) if a == "patch" else replace(
        p, trace=None
    ) if a == "trace" else replace(
        p, options=None
    ) if a == "options" else replace(
        p, head=None
    )

_prism_o = (lambda a: a, lambda b: b)

def odl(getter: Callable[[C], D], setter: Callable[[C, D], C]) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    # TODO: is there a better `None` prism?
    # Seems verbose
    return lens.Lens(getter, setter).Prism(lambda a: a, lambda a: a, ignore_none=True)

def _schema_o_setter(a: C, b: Optional[Schema]):
    return replace(a, schema = b)

schema_o = odl(lambda a: a.schema, _schema_o_setter)

def _request_body_o_setter(a: C, b: Optional[RequestBody]):
    return replace(a, requestBody = b)
    
request_body_o = odl(lambda a: a.requestBody, _request_body_o_setter)

def _content_o_setter(a: C, b: Optional[Mapping[str, MediaType]]):
    return replace(a, content = b)

content_o = odl(lambda a: a.content, _content_o_setter)

def _paths_o_setter(a: C, b: Optional[Mapping[str, PathItem]]):
    return replace(a, paths = b)

paths_o = odl(lambda a: a.paths, _paths_o_setter)

def _headers_o_setter(a: C, b: Optional[Mapping[str, Union[Header, Reference]]]):
    return replace(a, headers = b)   

headers_o = odl(lambda a: a.headers, _headers_o_setter)

def _responses_o(a: C, b: Responses):
    return replace(a, responses = b)      

responses_o = odl(lambda a: a.responses, _responses_o)

def _parameters_o_setter(a: C, b: Optional[Sequence[Parameter]]):
    return replace(a, parameters = b)      

parameters_o = odl(lambda a: a.parameters, _parameters_o_setter)

def _get_o_setter(a: C, b: Optional[Operation]):
    return replace(a, get = b)          

def _post_o_setter(a: C, b: Optional[Operation]):
    return replace(a, post = b)

def _put_o_setter(a: C, b: Optional[Operation]):
    return replace(a, put = b) 

def _delete_o_setter(a: C, b: Optional[Operation]):
    return replace(a, delete = b)

def _options_o_setter(a: C, b: Optional[Operation]):
    return replace(a, options = b)

def _head_o_setter(a: C, b: Optional[Operation]):
    return replace(a, head = b)

def _patch_o_setter(a: C, b: Optional[Operation]):
    return replace(a, patch = b)

def _trace_o_setter(a: C, b: Optional[Operation]):
    return replace(a, trace = b)

def operation_o(m: str):
    return {
        "get": odl(lambda a: a.get, _get_o_setter),
        "post": odl(lambda a: a.post, _post_o_setter),
        "put": odl(lambda a: a.put, _put_o_setter),
        "delete": odl(lambda a: a.delete, _delete_o_setter),
        "options": odl(lambda a: a.options, _options_o_setter),
        "head": odl(lambda a: a.head, _head_o_setter),
        "patch": odl(lambda a: a.patch, _patch_o_setter),
        "trace": odl(lambda a: a.trace, _trace_o_setter),
    }[m]

def oll(focus: int) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    '''Focus on a specific index in an array if it exists 
    '''
    return lens.Items().Filter(lambda a: a[0] == focus)[1]
    

def match_urls(protocol: str, host: str, o: OpenAPIObject) -> Sequence[str]:
    """Finds server URLs that match a given protocol and host.

    Arguments:
        protocol {str} -- like http or https
        host {str} -- like api.foo.com
        o {OpenAPIObject} -- schema from which the server URLs are taken
    
    Returns:
        A list of URLs that match the OpenAPI spec.
    """
    servers = o.servers
    if servers is None:
        return []
    #### if the openapi server url has no scheme,
    #### we ignore the incoming scheme and treat the url as the host
    #### not sure if this is the right decision,
    #### but it deals with a realistic outcome in certain schemas
    return [server.url
        for server in servers
        if ((urlparse(server.url).scheme in ['', None]) or (urlparse(server.url).scheme == protocol))
            and ((server.url.split('/')[0] == host) or (urlparse(server.url).netloc == host))]

def get_component_from_ref(o: OpenAPIObject, d: str, accessor: Callable[[Components], Optional[Mapping[str, Union[Reference, C]]]], getter: Callable[[OpenAPIObject, Union[C, Reference]], Optional[C]]) -> Optional[C]:
    return lens.F(
        lambda a: a.components
    ).F(
        lambda a: a if a is None else accessor(a)
    ).F(
        lambda a: a if a is None else getter(o, a[d])
    ).get()(o)

def ref_name(r: Reference) -> str:
    return r._ref.split("/")[3]

def internal_get_component(f: Callable[[OpenAPIObject, str], Optional[C]]) -> Callable[[OpenAPIObject, Union[C,  Reference]], Optional[C]]:
    def _internal_get_component(o: OpenAPIObject, i: Union[C,  Reference]) -> Optional[C]:
        return (f(o, ref_name(i)) if isinstance(i, Reference) else i) if i is not None else None
    return _internal_get_component

# TODO: it seems that there is a bug in pyright that makes get_component_from_ref return a Union[Reference, Schema] instead of Schema
# need to do a cast of the output value to Schema (instead of the union) to fix that
# same for functions below
def get_schema_from_ref(o: OpenAPIObject, d: str) -> Optional[Schema]:
    out = get_component_from_ref(o, d, lambda a: a.schemas, internal_get_schema_from_ref)
    return None if out is None else cast(Schema, out)

internal_get_schema_from_ref = internal_get_component(get_schema_from_ref)

def get_parameter_from_ref(o: OpenAPIObject, d: str) -> Optional[Parameter]:
    out = get_component_from_ref(o, d, lambda a: a.parameters, internal_get_parameter_from_ref)
    return None if out is None else cast(Parameter, out)

internal_get_parameter_from_ref = internal_get_component(get_parameter_from_ref)

def get_request_body_from_ref(o: OpenAPIObject, d: str) -> Optional[RequestBody]:
    out = get_component_from_ref(o, d, lambda a: a.requestBodies, internal_get_request_body_from_ref)
    return None if out is None else cast(RequestBody, out)

internal_get_request_body_from_ref = internal_get_component(get_request_body_from_ref)

def get_response_from_ref(o: OpenAPIObject, d: str) -> Optional[Response]:
    out = get_component_from_ref(o, d, lambda a: a.responses, internal_get_response_from_ref)
    return None if out is None else cast(Response, out)

internal_get_response_from_ref = internal_get_component(get_response_from_ref)


def schema_prism(oai: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Prism(
        lambda s: get_schema_from_ref(oai, ref_name(s)) if isinstance(s, Reference) else s,
        lambda a: a,
        ignore_none=True
    )

def change_ref(j: Reference) -> Reference:
    return Reference(
        _x=None,
        _ref='#/definitions/%s' % ref_name(j),
    )

def change_refs(j: Schema) -> Schema:
    return replace(j,
        additionalProperties=None if j.additionalProperties is None
            else j.additionalProperties
            if isinstance(j.additionalProperties, bool)
            else change_ref(j.additionalProperties)
            if isinstance(j.additionalProperties, Reference)
            else change_refs(j.additionalProperties),
        items=None if j.items is None
            else change_ref(j.items)
            if isinstance(j.items, Reference)
            else change_refs(j.items)
            if isinstance(j.items, Schema)
            else [change_ref(item) if isinstance(item, Reference) else change_refs(item) for item in j.items],
        properties=None if j.properties is None
            else { k: change_ref(v) if isinstance(v, Reference) else change_refs(v) for k, v in j.properties.items() }
    )

def make_definitions_from_schema(o: OpenAPIObject) -> Mapping[str, Any]:
    return {
        a: convert_from_openapi(change_ref(b) if isinstance(b, Reference) else change_refs(b)) for a, b in o.components.schemas.items()
    } if o.components and o.components.schemas else {}

def find_relevant_path(m: str, a: Sequence[str], p: PathItem) -> PathItem:
  return p if len(a) == 0 else find_relevant_path(
        m,
        a[1:],
        p if a[0] == m else omit_method_from_path_item(p, a[0]),
      )


def get_path_item_with_method(m: str, p: PathItem) -> PathItem:
    return find_relevant_path(m, all_methods, p)

def get_required_request_query_or_header_parameters_internal(header: bool, l: lenses.ui.BaseUiLens[S, T, X, Y], oai: OpenAPIObject, p: PathItem) -> Sequence[Parameter]:
    # TODO: really dirty cast
    # is this even the case
    # copied from unmock, but need to investigate
    return [replace(parameter,
                name=parameter.name if not header else parameter.name.lower(),
                schema=change_ref(parameter.schema) if isinstance(parameter.schema, Reference) else change_refs(parameter.schema) if parameter.schema is not None else Schema(**{**_SCHEMA_DEFAULT, '_type': 'string'})
                ) for parameter in l.Prism(
        lambda s: get_parameter_from_ref(oai, ref_name(s)) if isinstance(s, Reference) else s,
        lambda a : a,
        ignore_none=True
    ).Prism(
        lambda s : s if (s._in == ( "header" if header else "query")) and s.required else None,
        lambda a : a,
        ignore_none=True
    ).collect()(p) if parameter.schema is not None and (len(lens.Each().add_lens(schema_prism(oai)).collect()([parameter.schema])) > 0)]

def get_required_request_query_or_header_parameters(header: bool, req: Request, oai: OpenAPIObject, p: PathItem) -> Sequence[Parameter]:
    return [
        *get_required_request_query_or_header_parameters_internal(
            header,
            parameters_o.Each(),
            oai,
            p),
        *get_required_request_query_or_header_parameters_internal(
            header,
            operation_o(req.method.value).add_lens(
                parameters_o
            ).Each(),
            oai,
            p)
    ]

def get_required_request_body_schemas(
  req: Request,
  oai: OpenAPIObject,
  p: PathItem,
) -> Sequence[Union[Reference, Schema]]:
    return operation_o(req.method.value).add_lens(
        request_body_o
    ).Prism(
        lambda s : get_request_body_from_ref(oai, ref_name(s)) if isinstance(s, Reference) else s,
        lambda a: a,
        ignore_none=True
    ).add_lens(
        content_o
    ).Values().add_lens(
        schema_o
    ).add_lens(schema_prism(oai)).collect()(p)

def valid_schema(to_validate: Any, schema: Any) -> bool:
    try:
        jsonschema.validate(to_validate, schema)
        return True
    except Exception:
        # print(e)
        return False

def keep_method_if_required_request_body_is_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_request_body_is_present(p: PathItem) -> PathItem:
        out = p if (operation_from_string(p, req.method.value) is None) or (len([s for s in get_required_request_body_schemas(req, oai, p) if
                not valid_schema(req.bodyAsJson if req.bodyAsJson is not None else json.loads(req.body) if req.body is not None else '', {
                    # TODO: this line is different than the TS implementation
                    # because I think there is a logic bug there
                    # it should look like this line as we are not sure
                    # if the schema will be a reference or not
                    # perhaps I'm wrong in the assumption... only testing will tell...
                    # otherwise, change the TS implementation in unmock-js and delete this comment.
                    **convert_from_openapi(change_ref(s) if isinstance(s, Reference) else change_refs(s)),
                    'definitions': make_definitions_from_schema(oai),
                })]) == 0) else omit_method_from_path_item(p, req.method.value)
        return out
    return _keep_method_if_required_request_body_is_present

def keep_method_if_required_header_parameters_are_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_header_parameters_are_present(p: PathItem) -> PathItem:
        out = keep_method_if_required_query_or_header_parameters_are_present(True, req, oai, p)
        return out
    return _keep_method_if_required_header_parameters_are_present

def keep_method_if_required_query_parameters_are_present(
    req: Request,
    oai: OpenAPIObject,
) -> Callable[[PathItem], PathItem]:
    def _keep_method_if_required_query_parameters_are_present(p: PathItem) -> PathItem:
        out = keep_method_if_required_query_or_header_parameters_are_present(False, req, oai, p)
        return out
    return _keep_method_if_required_query_parameters_are_present

def _json_schema_from_required_parameters(parameters: Sequence[Parameter], oai: OpenAPIObject) -> Any:
    return {
        'type': 'object',
        'properties': {param.name: convert_from_openapi(param.schema) for param in parameters},
        'required': [param.name for param in parameters if param.required],
        'additionalProperties': True,
        'definitions': make_definitions_from_schema(oai),
    }


def keep_method_if_required_query_or_header_parameters_are_present(
    header: bool,
    req: Request,
    oai: OpenAPIObject,
    p: PathItem
) -> PathItem:
    return p if (operation_from_string(p, req.method.value) is None) or valid_schema(
        { k.lower(): v for k, v in req.headers.items() } if header else req.query,
        _json_schema_from_required_parameters(get_required_request_query_or_header_parameters(header, req, oai, p), oai)
    ) else omit_method_from_path_item(p, req.method.value)

def maybe_add_string_schema(s: Sequence[Union[Reference, Schema]]) -> Sequence[Union[Reference, Schema]]:
    return [Schema(**{**_SCHEMA_DEFAULT, '_type': "string"})] if len(s) == 0 else s

def discern_name(o: Optional[Parameter], n: str) -> Optional[Parameter]:
    return o if (o is None) or ((o.name == n) and (o._in == 'path')) else None

def internal_get_parameter_schemas(
    t: lenses.ui.BaseUiLens[S, T, X, Y],
    vname: str,
    path_item: PathItem,
    oas: OpenAPIObject,
) -> Sequence[Schema]:
    return t.Prism(
        lambda i: discern_name(get_parameter_from_ref(oas, ref_name(i)), vname) if isinstance(i, Reference) else discern_name(i, vname),
        lambda a: a,
        ignore_none=True
    ).add_lens(
        schema_o
    ).collect()(path_item)


def path_item_path_parameter_schemas(
    vname: str,
    path_item: PathItem,
    oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(
        parameters_o.Each(),
        vname,
        path_item,
        oas,
    )

def operation_path_parameter_schemas(
  vname: str,
  path_item: PathItem,
  operation: str,
  oas: OpenAPIObject,
) -> Sequence[Schema]:
    return internal_get_parameter_schemas(
        operation_o(
            operation.lower()
        ).add_lens(
            parameters_o
        ).Each(),
        vname,
        path_item,
        oas,
  )

def get_matching_parameters(
  vname: str,
  path_item: PathItem,
  operation: str,
  oas: OpenAPIObject,
) -> Sequence[Union[Schema, Reference]]:
    return maybe_add_string_schema([
        *path_item_path_parameter_schemas(vname, path_item, oas),
        *operation_path_parameter_schemas(vname, path_item, operation, oas)
    ])

def maybeJson(maybe: str) -> Any:
  try:
    return json.loads(maybe)
  except:
    return maybe

def path_parameter_match(
    part: str,
    vname: str,
    path_item: PathItem,
    operation: str,
    oas: OpenAPIObject,
) -> bool:
    '''Matches part of a path against a path parameter with name vname
    
    Arguments:
        part {str} -- part of a path, ie an id
        vname {str} -- name of a parameter
        path_item {PathItem} -- a path item maybe containing the parameter
        operation {MethodNames} -- the name of the operation to check in case the parameter is in the operation
        oas {OpenAPIObject} -- the schema to traverse to find definitions
    
    Returns:
        bool -- Did we get a match
    '''
    return reduce(lambda q, r: q and reduce(
        lambda a, b:
          a or
          valid_schema(b, {
            **convert_from_openapi(r),
            'definitions': make_definitions_from_schema(oas),
          }),
        list(set([part, maybeJson(part)])),
        False,
    ), get_matching_parameters(vname, path_item, operation, oas), True)

path_param_regex = re.compile(r'^\{[^}]+\}')

# TODO: this is a boolean nightmare
# what we want is
# lengths_are_equal AND (empty OR ((first_elt_equal OR (matches_regex AND matches_schema)) AND recursion))
def matches_internal(
    path: Sequence[str],
    path_item_key: Sequence[str],
    path_item: PathItem,
    operation: str,
    o: OpenAPIObject,
) -> bool:
    return (len(path) == len(path_item_key)) and (
        (len(path) == 0) or
        (((path[0] == path_item_key[0]) or
        ((path_param_regex.match(path_item_key[0]) is not None) and
            path_parameter_match(
                path[0],
                path_item_key[0][1:-1],
                path_item,
                operation,
                o,
            ))) and
            matches_internal(
                path[1:],
                path_item_key[1:],
                path_item,
                operation,
                o,
            )))

def matches(
  path: str,
  path_item_key: str,
  path_item: PathItem,
  method: str,
  oas: OpenAPIObject,
) -> bool:
  return matches_internal(
    [x for x in path.split("/") if x != ""],
    [x for x in path_item_key.split("/") if x != ""],
    path_item,
    method,
    oas,
  )

def get_first_method_internal_2(
  p: PathItem,
  n: str,
  m: Sequence[str],
  o: Optional[Operation],
) ->  Optional[Tuple[str, Operation]]:
    return (n, o) if o is not None else get_first_method_internal(m, p)
    

def get_first_method_internal(
  m: Sequence[str],
  p: PathItem,
) -> Optional[Tuple[str, Operation]]:
    return None if len(m) == 0 else get_first_method_internal_2(p, m[0], m[1:], operation_from_string(p, m[0]))

def get_first_method(p: PathItem) -> Optional[Tuple[str, Operation]]:
  return get_first_method_internal(all_methods, p)

def get_header_from_ref(
    o: OpenAPIObject,
    d: str
) -> Optional[Header]:
    out = get_component_from_ref(
        o,
        d,
        lambda a: a.headers,
        internal_get_header_from_ref,
    )
    return None if out is None else cast(Header, out)

internal_get_header_from_ref = internal_get_component(get_header_from_ref)

def use_if_header_last_mile(
  p: Parameter,
  r: Optional[Schema],
) -> Optional[Tuple[str, Schema]]:
  return None if r is None else (p.name, r if r is not None else Schema(**{**_SCHEMA_DEFAULT, '_type': "string"}))

def use_if_header(
  o: OpenAPIObject,
  p: Parameter,
) -> Optional[Tuple[str, Schema]]:
    return None if p._in != "header" else use_if_header_last_mile(
        p,
        Schema(**{**_SCHEMA_DEFAULT, '_type':"string"}) if p.schema is None else
            get_schema_from_ref(o, ref_name(p.schema)) if isinstance(p.schema, Reference) else
                p.schema
    )

def parameter_schema(o: OpenAPIObject) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return lens.Iso(lambda a: (use_if_header(o, a), a), lambda b: b[1])[0].Prism(*_prism_o, ignore_none=True)


def cut_path(paths: Sequence[str], path: str) -> str:
    return path if len(paths) == 0 else path[len(paths[0]):] if path[:len(paths[0])] == paths[0] else cut_path(paths[1:], path)

def remove_trailing_slash(s: str) -> str:
    return s if len(s) == 0 else s[:-1] if s[-1] == "/" else s

def truncate_path(
  path: str,
  o: OpenAPIObject,
  i: Request,
) -> str:
    return cut_path(
        [remove_trailing_slash(urlparse(u).path) for u in match_urls(i.protocol.value, i.host, o)],
        path,
    )

def matcher(req: Request, r: Mapping[str, OpenAPIObject]) -> Mapping[str, OpenAPIObject]:
    def _path_item_modifier(oai: OpenAPIObject) -> Callable[[PathItem], PathItem]:
        def __path_item_modifier(path_item: PathItem) -> PathItem:
            return reduce(lambda a, b: b(a), [
                    keep_method_if_required_header_parameters_are_present(req, oai),
                    keep_method_if_required_query_parameters_are_present(req, oai),
                    keep_method_if_required_request_body_is_present(req, oai),
                ], get_path_item_with_method(req.method.value, path_item))
        return __path_item_modifier
    def _oai_modifier(oai: OpenAPIObject) -> OpenAPIObject:
        return paths_o.Values().modify(
            _path_item_modifier(oai)
        )(replace(oai,
                    paths={n: o for n, o in oai.paths.items() if matches(
                    truncate_path(req.pathname, oai, req),
                    n,
                    o,
                    req.method.value,
                    oai,
                  )}
        ))
    return lens.Values().modify(
        _oai_modifier
    )({
        k: v for k, v in r.items() if len(match_urls(req.protocol.value, req.host, v)) > 0
    })

def _first_or_none(l: Sequence[C]) -> Optional[C]:
    return None if len(l) == 0 else l[0]

def headers_schemas_from_operation(schema: OpenAPIObject, operation: Operation) -> Sequence[Schema]:
    return parameters_o.Each().Prism(
        lambda s: s if isinstance(s, Reference) else get_parameter_from_ref(schema, ref_name(s)) if isinstance(s, Reference) else None,
        lambda a: a,
        ignore_none=True
    ).Iso(
        lambda a: (use_if_header(schema, a), a),
        lambda b: b[1]
    )[0].Prism(*_prism_o, ignore_none=True).collect()(operation)

def make_lens_to_response_starting_from_operation(
    schema: OpenAPIObject,
    code: str,
    l: lenses.ui.BaseUiLens[S, T, X, Y] = lens
) -> lenses.ui.BaseUiLens[S, T, X, Y]:
    return l.add_lens(responses_o).Prism(
        lambda s: s if isinstance(s, Response) else 
            get_response_from_ref(schema, ref_name(s)) if isinstance(s, Reference) else
            None,
        lambda a: a,
        ignore_none=True
    )

def string_header(n: str, o: Optional[Header]) -> Optional[Tuple[str, Header]]:
    return None if o is None else (n, o)

def header_schemas_from_response(
  schema: OpenAPIObject,
  operation: Operation,
  code: str
) -> Sequence[Schema]:
    return make_lens_to_response_starting_from_operation(
      schema, code
    ).add_lens(headers_o).Items().Prism(
      lambda a: string_header(a[0], get_header_from_ref(schema, ref_name(a[1]))) if isinstance(a[1], Reference) else (a[0], a[1]),
      lambda a: a,
      ignore_none=True
    ).Iso(
      lambda a: replace(a[1], name=a[0], _in='header'),
      lambda a: (a.name, a)
    ).add_lens(parameter_schema(schema)).collect()(operation)

def body_from_response(
  schema: OpenAPIObject,
  operation: Operation,
  code: str
) -> Optional[Union[Reference, Schema]]:
  return _first_or_none(
    make_lens_to_response_starting_from_operation(
          schema, code
    ).add_lens(
        content_o
    ).Items().add_lens(
        oll(0)
    )[1].add_lens(
        schema_o
    ).collect()([operation]))


########################
#### FAKER
########################
import random
from faker import Faker
fkr = Faker()
_LO = -99999999
_HI = 99999999

# to prevent too-nested objects
def sane_depth(n):
    return max([0, 3-n])

def fake_object(schema: Any, top_schema: Any, depth: int) -> Any:
    addls = {} if 'additionalProperties' not in schema else {k:v for k,v in [(fkr.name(), random.random() if (type(schema['additionalProperties']) == type(True)) and (schema['additionalProperties'] == True) else faker(schema['additionalProperties'], top_schema, depth)) for x in range(random.randint(0, 4))]}
    properties = [] if 'properties' not in schema else [x for x in schema['properties'].keys()]
    random.shuffle(properties)
    properties = [] if len(properties) == 0 else properties[:min([sane_depth(depth), random.randint(0, len(properties) - 1 )])]
    properties = list(set(([] if 'required' not in schema else schema['required']) + properties))
    return {
        **addls,
        **{ k: v for k,v in [(p, faker(schema['properties'][p], top_schema, depth)) for p in properties]}
    }

def fake_array(schema: Any, top_schema: Any, depth: int) -> Any:
    mn = 0 if 'minItems' not in schema else schema['minItems']
    mx = 100 if 'maxItems' not in schema else schema['maxItems']
    return [] if 'items' not in schema else [faker(x, top_schema, depth) for x in schema['items']] if type(schema['items']) is type([]) else [faker(schema['items'], top_schema, depth) for x in range(random.randint(mn, mx))]

def fake_any_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return faker(random.choice(schema["anyOf"]), top_schema, depth)

def fake_all_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return reduce(lambda a, b: { **a, **b}, [faker(x, top_schema, depth) for x in schema["allOf"]], {})

def fake_one_of(schema: Any, top_schema: Any, depth: int) -> Any:
    return faker(random.choice(schema["oneOf"]), top_schema, depth)

# TODO - make this work
def fake_not(schema: Any, top_schema: Any, depth: int) -> Any:
    return {}

# TODO - make this not suck
def fake_string(schema: Any) -> str:
    return random.choice(schema['enum']) if 'enum' in schema else fkr.name()

def fake_boolean(schema: Any) -> bool:
    return random.choice(schema['enum']) if 'enum' in schema else True if random.random() > 0.5 else False

# TODO: add exclusiveMinimum and exclusiveMaximum
def fake_integer(schema: Any) -> int:
    mn = _LO if 'minimum' not in schema else schema['minimum']
    mx = _HI if 'maximum' not in schema else schema['maximum']
    return random.choice(schema['enum']) if 'enum' in schema else random.randint(mn, mx)

def fake_ref(schema: Any, top_schema, depth):
    name = schema['$ref'].split('/')[2]
    return faker(top_schema['definitions'][name], top_schema, depth)

def fake_null(schema: Any) -> None:
    return None

def fake_number(schema: Any) -> float:
    mn = _LO if 'minimum' not in schema else schema['minimum']
    mx = _HI if 'maximum' not in schema else schema['maximum']
    return random.choice(schema['enum']) if 'enum' in schema else (random.random() * (mx - mn)) + mn

def faker(schema: Any, top_schema: Any, depth: int) -> Any:
    depth += 1
    return fake_array(
        schema, top_schema, depth
    ) if ('type' in schema) and (schema["type"] == "array") else fake_any_of(
        schema, top_schema, depth
    ) if "anyOf" in schema else fake_all_of(
        schema, top_schema, depth
    ) if "allOf" in schema else fake_one_of(
        schema, top_schema, depth
    ) if "oneOf" in schema else fake_not(
        schema, top_schema, depth
    ) if "not" in schema else fake_ref(
        schema, top_schema, depth
    ) if "$ref" in schema else fake_object(
        schema, top_schema, depth
    ) if ("type" not in schema) or (schema["type"] == "object") else fake_string(
        schema
    ) if schema["type"] == "string" else fake_integer(
        schema
    ) if schema["type"] == "integer" else fake_boolean(
        schema
    ) if schema["type"] == "boolean" else fake_null(
        schema
    ) if schema["type"] == "null" else fake_number(
        schema
    ) if schema["type"] == "number" else {}


import json
import logging
import os
from typing import cast

import yaml
from http_types import Request, Response, HttpExchange, HttpExchangeWriter, RequestBuilder, ResponseBuilder

from meeshkan.schemabuilder.update_mode import UpdateMode
from ...schemabuilder.builder import BASE_SCHEMA, update_openapi

logger = logging.getLogger(__name__)


class DataCallback:
    def log(self, request: Request, response: Response):
        raise NotImplementedError()


class RequestLoggingCallback():
    _log_dir: str
    _specs_dir: str

    def __init__(self, log_dir, specs_dir, update_mode, append=True):
        self._log_dir = log_dir
        self._specs_dir = specs_dir
        self._update_mode = update_mode
        self._append = append

        self._logs = {}
        self._specs = {}

        if not os.path.exists(self._log_dir):
            os.makedirs(self._log_dir)
        if not os.path.exists(self._specs_dir):
            os.makedirs(self._specs_dir)

    def log(self, request: Request, response: Response):
        RequestBuilder.validate(request)
        ResponseBuilder.validate(response)

        host = request.host
        reqres = HttpExchange(request=request, response=response)
        if not host in self._logs:
            log_file = os.path.join(self._log_dir, '{}.jsonl'.format(host))
            if self._append and os.path.exists(log_file):
                self._logs[host] = open(log_file, 'a')
            else:
                self._logs[host] = open(log_file, 'w')

        HttpExchangeWriter(self._logs[host]).write(reqres)
        self._logs[host].write('\n')
        self._logs[host].flush()

        logger.debug('Updated logs for host %s', host)

        if self._update_mode:
            spec_file = os.path.join(self._specs_dir, '{}_{}.yaml'.format(host, self._update_mode.name.lower()))

            if not host in self._specs:
                if os.path.exists(spec_file) and self._append:
                    with open(spec_file, 'r') as f:
                        self._specs[host] = yaml.load(f)
                else:
                    self._specs[host] = BASE_SCHEMA

            self._specs[host] = update_openapi(self._specs[host], reqres, self._update_mode)

            with open(spec_file, 'w') as f:
                yaml.dump(self._specs[host], f)

            logger.debug('Updated scheme for host %s', host)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for f in self._logs.values():
            f.close()



from meeshkan.schemabuilder.update_mode import UpdateMode
from meeshkan.schemabuilder.defaults import _SCHEMA_DEFAULT
from meeshkan.schemabuilder.param import ParamBuilder, SchemaParameters
from http_types import RequestBuilder
from openapi_typed_2 import Parameter, Schema
from typing import cast
from hamcrest import *

req = RequestBuilder.from_url(
    "https://petstore.swagger.io/v1/pets?id=32&car=ferrari")


def test_build_new_query():
    query = req.query
    schema_query = ParamBuilder('query').build(query, UpdateMode.GEN)
    assert_that(schema_query, has_length(2))

    query_parameter = get_query_parameter("id", schema_query)

    assert query_parameter.name == 'id'
    assert query_parameter._in == 'query'
    assert query_parameter.schema._type == 'string'


required_query_parameter = Parameter(description=None,
                                     deprecated=None,
                                     allowEmptyValue=None,
                                     style=None,
                                     explode=None,
                                     allowReserved=None,
                                     content=None,
                                     example=None,
                                     examples=None,
                                     _x=None,
                                     name='key',
                                     required=True,
                                     schema=Schema(**{ **_SCHEMA_DEFAULT, '_type': 'string'}),
                                     _in='query')


def get_query_parameter(name: str, parameters: SchemaParameters) -> Parameter:
    matching = [param for param in parameters if param.name == name][0]
    return cast(Parameter, matching)


def test_update_query_should_update_required():
    query = req.query
    updated_query = ParamBuilder('query').update(query, UpdateMode.GEN, [required_query_parameter])
    assert_that(updated_query, has_length(3))

    updated_for_key = get_query_parameter("key", updated_query)
    assert not updated_for_key.required



"""OpenAPI schema building operations. Uses request-response pairs as input.
"""
from .builder import *
from . import builder
from .update_mode import *
from . import update_mode

__all__ = [*builder.__all__, *update_mode.__all__]



import copy
from sys import path
from meeshkan.gen.generator import parameters_o
from .update_mode import UpdateMode
from functools import reduce
from dataclasses import replace
from typing import Any, List, Sequence, Iterable, AsyncIterable, cast, Tuple, Optional, Union, TypeVar, Type, Mapping
from urllib.parse import urlunsplit
from collections import defaultdict
import json

from http_types import HttpExchange as HttpExchange
from openapi_typed_2 import Info, Paths, MediaType, Header, OpenAPIObject, PathItem, Response, Operation, Parameter, Reference, Server, Responses
from typeguard import check_type  # type: ignore

from .operation import operation_from_string, new_path_item_at_operation
from ..logger import get as getLogger
from .media_types import infer_media_type_from_nonempty, build_media_type, update_media_type, MediaTypeKey, update_text_schema
from .paths import find_matching_path, RequestPathParameters
from .param import ParamBuilder
from .servers import normalize_path_if_matches
from .result import BuildResult

_DEFAULT_PATH_ITEM = {'servers': None, 'parameters': None, 'get': None, 'put': None, 'post': None, 'delete': None, 'options': None, 'head': None, 'patch': None, 'trace': None, '_ref': None, '_x': None} 

logger = getLogger(__name__)

__all__ = ['build_schema_batch', 'build_schema_online', 'update_openapi']

def build_response_content(exchange: HttpExchange, mode: UpdateMode) -> Optional[Tuple[MediaTypeKey, MediaType]]:
    """Build response content schema from exchange.

    Arguments:
        request {HttpExchange} -- Http exchange.

    Returns:
        Optional[Tuple[str, MediaType]] -- None for empty body, tuple of media-type key and media-type otherwise.
    """
    body = exchange.response.body

    if body == '':
        return None

    media_type_key = infer_media_type_from_nonempty(body)

    media_type = build_media_type(exchange, mode, type_key=media_type_key)

    return (media_type_key, media_type)


def build_response(request: HttpExchange, mode: UpdateMode) -> Response:
    """Build new response object from request response pair.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        request {HttpExchange} -- Request-response pair.

    Returns:
        Response -- OpenAPI response object.
    """
    # TODO Headers and links
    content_or_none = build_response_content(request, mode)

    if content_or_none is None:
        return Response(
            description="Response description",
            headers={},
            links={},
            content=None,
            _x=None
        )

    media_type_key, media_type = content_or_none
    content = {
        media_type_key: media_type
    }

    return Response(
        description="Response description",
        headers={},
        content=content,
        links={},
        _x=None
    )


def update_response(response: Response, mode: UpdateMode, exchange: HttpExchange) -> Response:
    """Update response object.

    Response reference: https://swagger.io/specification/#responseObject

    Arguments:
        response {Response} -- Existing response object.
        exchange {HttpExchange} -- Request-response pair.

    Returns:
        Response -- Updated response object.
    """
    # TODO Update headers and links
    response_body = exchange.response.body

    if response_body == '':
        # No body, do not do anything
        # TODO How to mark empty body as a possible response if non-empty responses exist
        return response

    media_type_key = infer_media_type_from_nonempty(response_body)

    ################
    ### HEADERS
    ################
    response_headers = response.headers if response.headers is not None else {}
    # TODO: accommodate array headers (currently cuts them off)
    useable_headers: Mapping[str, str] = { k: v for k,v in exchange.response.headers.items() if k not in ['content-type', 'content-length', 'Content-Type', 'Content-Length'] and isinstance(v, str)}
    new_headers = {
            **response_headers,
            **{
            k: Header(
                schema=update_text_schema(v, mode, schema=response.headers.get(k, None))) for k, v in useable_headers.items()
            }
        }

    #############
    ### CONTENT
    #############
    response_content = response.content
    if response_content is not None and media_type_key in response_content:
        # Need to update existing media type
        existing_media_type = response_content[media_type_key]
        media_type = update_media_type(
            exchange=exchange, mode=mode, type_key=media_type_key, media_type=existing_media_type)
    else:
        media_type = build_media_type(
            exchange=exchange, mode=mode, type_key=media_type_key)

    new_content: Mapping[str, MediaType] = {media_type_key: media_type}
    response_content = {**response_content, **new_content }
    return replace(response, headers=new_headers, content=response_content)


def build_operation(exchange: HttpExchange, mode: UpdateMode) -> Operation:
    """Build new operation object from request-response pair.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Operation object.
    """
    response = build_response(exchange, mode)
    code = str(exchange.response.statusCode)

    request_query_params = exchange.request.query
    request_header_params = exchange.request.headers
    schema_query_params = ParamBuilder('query').build(request_query_params, mode)
    schema_header_params = ParamBuilder('header').build(request_header_params, mode)

    operation = Operation(
        summary="Operation summary",
        description="Operation description",
        operationId="id",
        responses={code: response},
        parameters=[*schema_query_params, *schema_header_params])
    return operation


def update_operation(operation: Operation, request: HttpExchange, mode: UpdateMode) -> Operation:
    """Update OpenAPI operation object.

    Operation reference: https://swagger.io/specification/#operationObject

    Arguments:
        operation {Operation} -- Existing Operation object.
        request {HttpExchange} -- Request-response pair

    Returns:
        Operation -- Updated operation
    """
    responses = operation.responses  # type: Responses
    response_code = str(request.response.statusCode)
    response: Response
    if response_code in responses:
        # Response exists
        existing_response = responses[response_code]
        # TODO: why do we have this reference check?
        if not isinstance(existing_response, Reference):
            response = update_response(existing_response, mode, request)
        else:
            raise ValueError('Meeshkan is not smart enough to build responses from references yet. Coming soon!')
    else:
        response = build_response(request, mode)

    # TODO: this is not right, we need to grab the references at some point!
    existing_parameters: Sequence[Parameter] = [x for x in operation.parameters if not isinstance(x, Reference)] if operation.parameters is not None else []
    request_query_params: Mapping[str, Union[str, Sequence[str]]] = request.request.query
    request_header_params: Mapping[str, Union[str, Sequence[str]]] = request.request.headers
    _updated_parameters = [*ParamBuilder('query').update(request_query_params,
            mode, existing_parameters),
            # TODO: can we avoid the cast below? it is pretty hackish
        *ParamBuilder('header').update(request_header_params,
            mode, existing_parameters)]
    updated_parameters: List[Parameter] = []
    for i, param in enumerate(_updated_parameters):
        if len([x for x in _updated_parameters[i+1:] if (x.name == param.name) and (x._in == param._in)]) == 0:
            updated_parameters.append(param)
    new_responses: Mapping[str, Response] = { response_code: response }
    return replace(operation, responses={ **operation.responses, **new_responses }, parameters = updated_parameters)


T = TypeVar('T')


def verify_not_ref(item: Union[Reference, Any], expected_type: Type[T]) -> T:
    assert not '$ref' in item, "Did not expect a reference"
    return cast(T, item)


def update_openapi(schema: OpenAPIObject, exchange: HttpExchange, mode: UpdateMode) -> OpenAPIObject:
    """Update OpenAPI schema with a new request-response pair.
    Does not mutate the input schema.

    Schema reference: https://swagger.io/specification/#oasObject

    Returns:
        OpenAPI -- Updated schema
    """
    request_method = exchange.request.method.value
    request_path = exchange.request.pathname

    serverz = [] if schema.servers is None else schema.servers

    normalized_pathname_or_none = normalize_path_if_matches(
        exchange.request, serverz)
    if normalized_pathname_or_none is None:
        normalized_pathname = request_path
    else:
        normalized_pathname = normalized_pathname_or_none

    schema_paths = schema.paths
    operation_candidate = build_operation(exchange, mode)
    path_match_result = find_matching_path(normalized_pathname, schema_paths, request_method, operation_candidate)

    request_path_parameters = {}
    make_new_paths = True
    if path_match_result is not None:
        # Path item exists for request path
        path_item, request_path_parameters, pathname_with_wildcard, pathname_to_be_replaced_with_wildcard = path_match_result.path, path_match_result.param_mapping, path_match_result.pathname_with_wildcard, path_match_result.pathname_to_be_replaced_with_wildcard
        # Create a wildcard if the mode is not replay/mixed
        if (pathname_to_be_replaced_with_wildcard is not None):
            if mode == UpdateMode.GEN:
                # the algorithm has updated the pathname, need to mutate
                # the schema paths to use the new and discard the old if the old exists
                # in the schema. it would not exist if we have already put a wildcard
                pointer_to_value = schema_paths[pathname_to_be_replaced_with_wildcard]
                schema_paths = { k: v for k, v in [(pathname_with_wildcard, pointer_to_value), *schema_paths.items()] if k != pathname_to_be_replaced_with_wildcard }
                parameters_to_assign_to_pathname_with_wildcard = [] if schema_paths[pathname_with_wildcard].parameters is None else schema_paths[pathname_with_wildcard].parameters
                for path_param in request_path_parameters.keys():
                    params = [x for x in parameters_to_assign_to_pathname_with_wildcard if not isinstance(x, Reference)]
                    if not (path_param in [x.name for x in params if x._in == 'path']):
                        parameters_to_assign_to_pathname_with_wildcard = [Parameter(
                            required=True,
                            _in='path',
                            name=path_param,
                        ), *parameters_to_assign_to_pathname_with_wildcard]
                schema_paths = {
                    **schema_paths,
                    pathname_with_wildcard: replace(schema_paths[pathname_with_wildcard], parameters=parameters_to_assign_to_pathname_with_wildcard)
                }
                make_new_paths = False # because we've already done it above
            else:
                # we are using recordings, so we shouldn't overwrite anything
                # we only add if it is not there yet
                if normalized_pathname not in schema_paths:
                    # TODO: merge with liens below?
                    path_item = PathItem(summary="Path summary",
                             description="Path description")
                    request_path_parameters = {}
    else:
        path_item = PathItem(summary="Path summary",
                             description="Path description")
        request_path_parameters = {}
        new_paths: Paths = { normalized_pathname: path_item }
    existing_operation = operation_from_string(path_item, request_method)
    if existing_operation is not None:
        # Operation exists
        operation = update_operation(existing_operation, exchange, mode)
    else:
        operation = operation_candidate

    path_item = new_path_item_at_operation(path_item, request_method, operation)
    new_paths = { path_match_result.pathname_with_wildcard if (mode == UpdateMode.GEN) and ((path_match_result is not None) and (path_match_result.pathname_with_wildcard is not None)) else normalized_pathname: path_item } if make_new_paths else {}
    new_server = Server(url=urlunsplit(
            [str(exchange.request.protocol.value), exchange.request.host, '', '', '']))
    return replace(
        schema,
        paths={ **schema_paths, **new_paths },
        servers=serverz if (new_server.url in [x.url for x in serverz]) else [*serverz, new_server]
    )


BASE_SCHEMA = OpenAPIObject(openapi="3.0.0",
                            info=Info(title="API title",
                                      description="API description",
                                      version="1.0",
                                      termsOfService=None,
                                      contact=None,
                                      _license=None,
                                      _x=None),
                            paths={},
                            externalDocs=None,
                            servers=None,
                            security=None,
                            tags=None,
                            components=None,
                            _x=None)


async def build_schema_async(async_iter: AsyncIterable[HttpExchange],  mode: UpdateMode, starting_spec: OpenAPIObject) -> AsyncIterable[BuildResult]:
    schema = starting_spec
    async for exchange in async_iter:
        schema = update_openapi(schema, exchange, mode)
        yield BuildResult(openapi=schema)


def build_schema_online(requests: Iterable[HttpExchange], mode: UpdateMode, base_schema: OpenAPIObject = BASE_SCHEMA) -> OpenAPIObject:
    """Build OpenAPI schema by iterating request-response pairs.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {Iterator[HttpExchange]} -- Iterator of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """

    # Iterate over all request-response pairs in the iterator, starting from
    # BASE_SCHEMA
    schema = reduce(lambda schema, req: update_openapi(
        schema, req, mode), requests, base_schema)

    return schema


def build_schema_batch(requests: List[HttpExchange], mode: UpdateMode = UpdateMode.GEN, base_schema: OpenAPIObject = BASE_SCHEMA) -> OpenAPIObject:
    """Build OpenAPI schema from a list of request-response object.

    OpenAPI object reference: https://swagger.io/specification/#oasObject

    Arguments:
        requests {List[HttpExchange]} -- List of request-response pairs.

    Returns:
        OpenAPI -- OpenAPI object.
    """
    return build_schema_online(iter(requests), mode, base_schema=base_schema)



from openapi_typed_2 import PathItem, Operation, Optional
from typing import Callable, Mapping
from dataclasses import replace

def operation_from_string(p: PathItem, s: str) -> Optional[Operation]:
    lexicon: Mapping[str, Optional[Operation]] = {
            "get": p.get,
            "post": p.post,
            "put": p.put,
            "delete": p.delete,
            "options": p.options,
            "head": p.head,
            "patch": p.patch,
            "trace": p.trace,
        }
    if s not in lexicon:
        return None
    return lexicon[s]

def new_path_item_at_operation(p: PathItem, s: str, o: Operation) -> PathItem:
    lexicon = {
            "get": {'get': o},
            "post": {'post': o},
            "put": {'put': o},
            "delete": {'delete': o},
            "options": {'options': o},
            "head": {'head': o},
            "patch": {'patch': o},
            "trace": {'trace': o}
        }
    if s not in lexicon:
        return p
    return replace(p, **lexicon[s])


import json
import logging
from meeshkan.server.server.rest import rest_middleware_manager
import os
import yaml
import random
from faker import Faker
from typing import cast, Mapping, Union, Sequence, Tuple
from openapi_typed_2 import convert_from_openapi, OpenAPIObject, Reference, Schema, convert_to_openapi
from ...gen.generator import get_response_from_ref, matcher, faker, change_ref, change_refs, ref_name
from http_types import HttpMethod
fkr = Faker()

from http_types import Request, Response

logger = logging.getLogger(__name__)

class ResponseMatcher:
    _schemas: Mapping[str, OpenAPIObject]
    def __init__(self, specs_dir):
        schemas: Sequence[str] = []
        if not os.path.exists(specs_dir):
            logging.info('OpenAPI schema directory not found %s', specs_dir)
        else:
            schemas = [s for s in os.listdir(specs_dir) if s.endswith('yml') or s.endswith('yaml') or s.endswith('json')]
        specs: Sequence[Tuple[str, OpenAPIObject]] = []
        for schema in schemas:
            with open(os.path.join(specs_dir, schema), encoding='utf8') as schema_file:
                # TODO: validate schema?
                specs = [*specs, (schema, convert_to_openapi((json.loads if schema.endswith('json') else yaml.safe_load)(schema_file.read())))]
        self._schemas = { k: v for k, v in specs}

    def match_error(self, msg: str, req: Request) -> Response:
        return self.default_response('%s. Here is the full request: host=%s, path=%s, method=%s.' % (msg, req.host, req.path, req.method.value))

    def default_response(self, msg):
        json_resp = {'message': msg}
        return Response(statusCode=500, body=json.dumps(json_resp), bodyAsJson=json_resp,
                        headers={}, timestamp=None)


    def get_response(self, request: Request) -> Response:
        # TODO: tight coupling here
        # try to decouple...
        schemas = rest_middleware_manager.spew(request, self._schemas)
        match = matcher(request, schemas)
        if len(match) == 0:
            return self.match_error('Could not find a open API schema for the host %s.' % request.host, request)
        match_keys = [x for x in match.keys()]
        random.shuffle(match_keys)
        name = match_keys[0]
        path_error = 'Could not find a path %s on hostname %s.' % (request.path, request.host)
        method_error = 'Could not find a method %s for path %s on hostname %s.' % (request.method.value, request.path, request.host)
        if match[name].paths is None:
            return self.match_error(path_error, request)
        if len(match[name].paths.items()) == 0:
            return self.match_error(path_error, request)
        path_candidates = [x for x in match[name].paths.values()]
        random.shuffle(path_candidates)
        path_candidate = path_candidates[0]
        method = {
            "get": path_candidate.get,
            "post": path_candidate.post,
            "put": path_candidate.put,
            "delete": path_candidate.delete,
            "options": path_candidate.options,
            "head": path_candidate.head,
            "patch": path_candidate.patch,
            "trace": path_candidate.trace,
        }.get(str(request.method.value), None)
        if request.method.value is None:
            return self.match_error(method_error, request)
        responses_error = 'While a stub for a specification exists for this endpoint, it contains no responses. That usually means the schema is corrupt or it has been constrained too much (ie asking for a 201 response when it only has 200 and 400).'
        if method.responses is None:
            return self.match_error(responses_error, request)
        potential_responses = [r for r in method.responses.items()]
        random.shuffle(potential_responses)
        if len(potential_responses) == 0:
            return self.match_error(responses_error, request)
        response = potential_responses[0]
        response_1 = response[1]
        response_1 = get_response_from_ref(match[name], ref_name(response_1)) if isinstance(response_1, Reference) else response_1
        if response_1 is None:
            return self.match_error(responses_error, request)
        headers: Mapping[str, Union[str, Sequence[str]]] = {}
        if response_1.headers is not None:
            # TODO: can't handle references yet, need to fix
            headers = {} # { k: (faker(v['schema'], v['schema'], 0) if 'schema' in v else '***') for k,v in headers.items() }
        statusCode = int(response[0] if response[0] != 'default' else 400)
        if (response_1.content is None) or len(response_1.content.items()) == 0:
            return Response(statusCode=statusCode, body="", headers=headers, bodyAsJson=None, timestamp=None)
        mime_types = response_1.content.keys()
        if "application/json" in mime_types:
            content = response_1.content['application/json']
            if content.schema is None:
                return self.match_error('Could not find schema', request)
            schema = content.schema
            ct: Mapping[str, Union[str, Sequence[str]]] = { "Content-Type": "application/json" }
            new_headers: Mapping[str, Union[str, Sequence[str]]] = { **headers, **ct }
            if schema is None:
                return Response(
                    statusCode=statusCode,
                    body="",
                    bodyAsJson="",
                    headers=new_headers,
                    timestamp=None
                )
            to_fake = {
                **convert_from_openapi(change_ref(schema) if isinstance(schema, Reference) else change_refs(schema)),
                'definitions': { k: convert_from_openapi(change_ref(v) if isinstance(v, Reference) else change_refs(v)) for k,v in (match[name].components.schemas.items() if (name in match) and (match[name].components is not None) and (match[name].components.schemas is not None) else [])}
            }
            bodyAsJson = faker(to_fake, to_fake, 0)
            return Response(
                statusCode=statusCode,
                body=json.dumps(bodyAsJson),
                bodyAsJson=bodyAsJson,
                headers=new_headers,
                timestamp=None
            )
        if "text/plain" in mime_types:
            return Response(
                statusCode=statusCode,
                body=fkr.sentence(),
                # TODO: can this be accomplished without a cast?
                headers=cast(Mapping[str, Union[str, Sequence[str]]], { **headers, "Content-Type": "text/plain" }),
                bodyAsJson=None,
                timestamp=None
            )
        return self.match_error('Could not produce content for these mime types %s' % str(mime_types), request)



import logging
from typing import Mapping
from openapi_typed_2 import OpenAPIObject, convert_to_openapi, convert_from_openapi
from http_types import Request
from http_types.utils import HttpExchangeWriter, ResponseBuilder, HttpExchange
from io import StringIO
import requests
import json
import copy

logger = logging.getLogger(__name__)


class RestMiddlewareManager:
    def __init__(self):
        self._endpoints = set([])


    def get(self):
        return list(self._endpoints)
    
    def clear(self, url=None):
        if not url:
            self._endpoints = set([])
        else:
            self._endpoints.remove(url)

    def add(self, url):
        self._endpoints.add(url)
    
    def spew(self, request: Request, schemas: Mapping[str, OpenAPIObject]) -> Mapping[str, OpenAPIObject]:
        _request = copy.deepcopy(request)
        _request.body = str(_request.body)
        cur_schemas = schemas
        req_io = StringIO()
        # TODO: this is hackish. is there a better way?
        HttpExchangeWriter(req_io).write(HttpExchange(request=_request, response=ResponseBuilder.from_dict(dict(statusCode=200,body='',headers={}))))
        # should only be one line... and why do we join with newline?
        req_io.seek(0)
        req = json.loads('\n'.join([x for x in req_io]))['request']
        cs = {k: convert_from_openapi(v) for k, v in cur_schemas.items()}
        for endpoint in self._endpoints:
            res = requests.post(endpoint, data=json.dumps({'request': req, 'schemas': cs }))
            cs = json.loads(res.text)
        out: Mapping[str, OpenAPIObject] = {str(k): convert_to_openapi(v) for k, v in cs.items()}
        return out


rest_middleware_manager = RestMiddlewareManager()



"""Code for working with parameters."""

from functools import reduce
from meeshkan.schemabuilder.json_schema import to_const
from meeshkan.schemabuilder.update_mode import UpdateMode
from openapi_typed_2 import convert_to_Schema, Parameter, Reference, Schema, convert_to_Schema
from .defaults import _SCHEMA_DEFAULT
from typing import Sequence, Union, cast, Sequence, List, Mapping
from genson import SchemaBuilder  # type: ignore
from dataclasses import replace

SchemaParameters = Sequence[Parameter]

# TODO: this won't work for anything other than simple schema
# maybe make more robust?
def unnest(d: Sequence[Union[Reference, Schema]]) -> Sequence[Union[Reference, Schema]]:
    return sum(
        [[x] if isinstance(x, Reference) or (x.oneOf is None) else unnest(x.oneOf) for x in d], [])
    

# TODO: extended this for all paramaters, but need to verify that
# there is not some subtle difference between query and header

class ParamBuilder:
    def __init__(self, _in: str):
        if _in not in ['header', 'query']:
            raise ValueError('A parameter cannot have an `in` value of %s.' % _in)
        self._in = _in

    def build(self, params: Mapping[str, Union[str, Sequence[str]]], mode: UpdateMode) -> SchemaParameters:
        """Build a list of parameters from request parameters.

        Arguments:
            params {Mapping[str, str]} -- Key-value map of params parameters and values.

        Returns:
            SchemaParameters -- OpenAPI list of parameters.
        """
        existing = []
        # as a defualt, we set nothing new to required
        # can change later if there is a compelling reason
        return self.update(params, mode, existing, set_new_as_required=False)


    def build_param(self, name: str, value: Union[str, Sequence[str]], required: bool, mode: UpdateMode) -> Parameter:
        """Build a new OpenAPI compatible parameter definition from parameter
        name and value.

        Arguments:
            name {str} -- Parameter name.
            value {Any} -- Parameter value.
            required {bool} -- Whether the parameter should be marked as required.
        Returns:
            Parameter -- [description]
        """
        # TODO: this may not be accurate if there are duplicate parameters
        # which is allowed in http
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        else:
            value = value
        out: Schema
        if (mode == UpdateMode.GEN):
            schema_builder = SchemaBuilder()
            schema_builder.add_object(value)
            out = convert_to_Schema(schema_builder.to_schema())
        else:
            out = Schema(**{**_SCHEMA_DEFAULT, 'oneOf': [convert_to_Schema(to_const(value))]})
        schema = out
        return Parameter(
            description=None,
            deprecated=None,
            allowEmptyValue=None,
            style=None,
            explode=None,
            allowReserved=None,
            content=None,
            example=None,
            examples=None,
            _x=None,
            name=name,
            schema=schema,
            required=required,
            _in=self._in
        )


    def _update_required(self, param: Parameter, required: bool) -> Parameter:
        if param.required == required:
            return param

        return Parameter(
            description=param.description,
            deprecated=param.deprecated,
            allowEmptyValue=param.allowEmptyValue,
            style=param.style,
            explode=param.explode,
            allowReserved=param.allowReserved,
            content=param.content,
            example=param.example,
            examples=param.examples,
            _x=param._x,
            name=param.name,
            _in=param._in,
            schema=param.schema,
            required=required
        )


    # TODO Fix types once openapi types are covariant
    def update(self, incoming_params: Mapping[str, Union[str, Sequence[str]]], mode: UpdateMode, existing: SchemaParameters, set_new_as_required=False) -> SchemaParameters:
        non_params: List[Parameter] = [
            param for param in existing if param._in != self._in]

        params: List[Parameter] = [
            param for param in existing if param._in == self._in]

        schema_params_names = frozenset(
            [param.name for param in params])

        request_param_names = frozenset(incoming_params.keys())

        # Parameters in request but not in schema
        new_param_names = request_param_names.difference(
            schema_params_names)

        # Parameters in schema but not in request
        missing_param_names = schema_params_names.difference(
            request_param_names)

        # Parameters in schema and in request
        shared_param_names = request_param_names.intersection(
            schema_params_names)

        new_params = [self.build_param(name=param_name,
                                            value=value,
                                            required=set_new_as_required,
                                            mode=mode)
                            for param_name in new_param_names
                            for value in (incoming_params[param_name],)]

        # TODO Update shared incoming_params parameter schema
        shared_params = [
            param if (mode == UpdateMode.GEN) else replace(param,
                schema=Schema(**{**_SCHEMA_DEFAULT, 'oneOf': unnest([*([param.schema] if param.schema is not None else []), convert_to_Schema(to_const(incoming_params[param.name]))])})
             ) for param in params if param.name in shared_param_names]

        missing_params = [
            self._update_required(param, required=False) for param in params if param.name in missing_param_names]

        # TODO: make sure cast is valid
        updated_params = [
            *new_params,
            *shared_params,
            *missing_params,
            *non_params
        ]
        return cast(SchemaParameters, updated_params)



import logging
import os

import click
import tornado.ioloop
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application

from .utils.routing import PathRouting, HeaderRouting, Routing
from .admin.views import RestMiddlewareView, RestMiddlewaresView, StorageView
from ..schemabuilder.update_mode import UpdateMode
from .admin.views import StorageView
from .proxy.proxy import RecordProxy
from .server.callbacks import callback_manager
from .server.response_matcher import ResponseMatcher
from .server.views import MockServerView
from .utils.data_callback import RequestLoggingCallback
from .utils.routing import PathRouting, HeaderRouting, Routing

LOG_CONFIG = os.path.join(os.path.dirname(__file__), 'logging.yaml')
logger = logging.getLogger(__name__)

def make_admin_app():
    return Application([
        (r'/admin/storage', StorageView),
        (r'/admin/middleware/rest/pregen', RestMiddlewaresView),
        (r'/admin/middleware/rest/pregen/(.+)', RestMiddlewareView),
    ])

def start_admin(port):
    app = make_admin_app()
    http_server = HTTPServer(app)
    http_server.listen(port)
    logger.info('Starting admin endpont on http://localhost:%s/admin', port)


@click.command()
@click.option('-p', '--port', default="8000", help='Server port.')
@click.option('-a', '--admin_port', default="8888", help='Admin server port.')
@click.option('-l', '--log_dir', default="./logs", help='API calls logs direcotry')
@click.option('-r', '--header_routing', is_flag=True, help='Whether to use a header based routing to a target host.')
@click.option('-s', '--specs_dir', default="./specs", help='Directory to store OpenAPI specs.')
@click.option("-m", "--mode", type=click.Choice(['GEN', 'REPLAY', 'MIXED'], case_sensitive=False),
              default=None, help="Spec building mode.")
def record(port, admin_port, log_dir, header_routing, specs_dir, mode):
    """
    Record http traffic to http-types format.
    """
    start_admin(admin_port)
    logger.info('Starting Meeshkan proxy on http://localhost:%s', port)
    logger.info('Spec generation mode is %s', mode.lower() if mode else 'disabled')
    with RequestLoggingCallback(log_dir=log_dir, specs_dir=specs_dir,
                                update_mode=UpdateMode[mode.upper()] if mode else None) as callback:
        server = RecordProxy(callback, HeaderRouting() if header_routing else PathRouting())
        server.listen(port)
        tornado.ioloop.IOLoop.instance().start()


class MeeshkanApplication(Application):
    response_matcher: ResponseMatcher
    router: Routing


def make_mocking_app(callback_path, specs_dir, router):
    app = MeeshkanApplication([
        (r'/.*', MockServerView)
    ])
    callback_manager.load(callback_path)

    app.response_matcher = ResponseMatcher(specs_dir)
    app.router = router
    return app


@click.command()
@click.option('-c', '--callback_path', default="./callbacks", help='Directory with configured callbacks.')
@click.option('-p', '--port', default="8000", help='Server port.')
@click.option('-a', '--admin_port', default="8888", help='Admin server port.')
@click.option('-s', '--specs_dir', default="./specs", help='Directory with OpenAPI schemas.')
@click.option('-r', '--header_routing', is_flag=True, help='Whether to use a path based routing to a target host.')
def mock(callback_path, admin_port, port, specs_dir, header_routing):
    start_admin(admin_port)
    app = make_mocking_app(callback_path, specs_dir, HeaderRouting() if header_routing else PathRouting())
    http_server = HTTPServer(app)
    http_server.listen(port)
    logger.info('Mock server is listening on http://localhost:%s', port)
    IOLoop.current().start()



from .prepare import ignore_warnings
ignore_warnings()

from .schemabuilder import *
from . import schemabuilder
from .gen import *
from . import gen

__all__ = [
    *schemabuilder.__all__,
    *gen.__all__
]



import warnings

def ignore_warnings():
    """Several of our deps have deprecation warnings
    # we ignore for now, but should look into."""
    ### This comes from faust """
    warnings.filterwarnings(
        action='ignore',
        category=DeprecationWarning,
        module=r'venusian'
    )
    ### This comes from lenses
    warnings.filterwarnings(
        action='ignore',
        category=DeprecationWarning,
        module=r'singledispatch_helpers'
    )



"""Code for writing builder results to file system.
"""
from openapi_typed_2 import OpenAPIObject, convert_from_openapi
from .result import BuildResult
import json
from ..logger import get as getLogger
from typing import cast
from pathlib import Path

LOGGER = getLogger(__name__)


OPENAPI_FILENAME = 'openapi.json'

def _ensure_dir_exists(path: Path) -> None:
    """Ensure directory at `path` exists. Does NOT create parent directories.

    Arguments:
        path {Path} -- Path to the directory to create.

    Raises:
        FileNotFoundError -- If the directory does not exist and its parents do not exist.
    """
    path.mkdir(parents=False, exist_ok=True)


def _resolve_path(directory: str) -> Path:
    """Return directory as absolute Path.

    Arguments:
        directory {str} -- Path to directory, possibly relative.

    Returns:
        Path -- Absolute Path.
    """
    path = Path(directory)
    return path.resolve()


def write_build_result(directory: str, result: BuildResult) -> None:
    """Write builder result to a directory.

    Arguments:
        directory {str} -- Directory where to write results, possibly relative.
        result {BuildResult} -- Builder result.
    """
    output_dir_path = _resolve_path(directory)
    LOGGER.info("Writing to folder %s.", str(output_dir_path))

    _ensure_dir_exists(output_dir_path)

    openapi_output = output_dir_path / OPENAPI_FILENAME

    openapi_object = result['openapi']

    schema_json = cast(str, json.dumps(convert_from_openapi(openapi_object)))

    with openapi_output.open('wb') as f:
        LOGGER.debug("Writing to: %s\n", str(openapi_output))
        f.write(schema_json.encode(encoding="utf-8"))



import logging

from tornado.web import RequestHandler

from ..server.storage import storage_manager
from ..server.rest import rest_middleware_manager
from tornado.web import RequestHandler
import json

logger = logging.getLogger(__name__)


class StorageView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE"]

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self):
        storage_manager.clear()


class RestMiddlewaresView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE", "GET"]

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self):
        rest_middleware_manager.clear()

    def get(self):
        self.write(json.dumps(rest_middleware_manager.get()))
        

class RestMiddlewareView(RequestHandler):
    SUPPORTED_METHODS = ["DELETE", "POST"]

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def delete(self, url):
        rest_middleware_manager.clear(url)

    def post(self, url):
        rest_middleware_manager.add(url)
        






import copy
from meeshkan.schemabuilder.update_mode import UpdateMode
import json
import re
from dataclasses import replace
from http_types import HttpExchange, Request, Response, RequestBuilder, HttpExchangeBuilder
from tests.schemabuilder.paths_test import PETSTORE_SCHEMA
from meeshkan.schemabuilder import build_schema_batch, update_openapi, build_schema_online
from meeshkan.schemabuilder.builder import BASE_SCHEMA
from meeshkan.schemabuilder.update_mode import UpdateMode
from meeshkan.schemabuilder.defaults import _SCHEMA_DEFAULT
from ..util import petstore_schema, read_recordings_as_request_response, POKEAPI_RECORDINGS_PATH
from openapi_typed_2 import OpenAPIObject, Operation, PathItem, Schema, Response as _Response
from typeguard import check_type
import pytest
from hamcrest import *

requests = read_recordings_as_request_response()
pokeapi_requests = read_recordings_as_request_response(POKEAPI_RECORDINGS_PATH)

expected_schema = replace(BASE_SCHEMA, paths = {
    '/user/repos': PathItem(
        get=Operation(
            responses={
                '200': _Response(description="description"),
                '403': _Response(description="description")
            }
        )
    )
})

PETSTORE_SCHEMA = petstore_schema()

@pytest.fixture
def schema():
    return build_schema_batch(requests, UpdateMode.GEN)


def test_schema_typechecks(schema):
    check_type('schema', schema, OpenAPIObject)

def test_schema_keys(schema):
    assert schema.openapi
    assert schema.info
    assert schema.paths

def test_paths(schema):
    paths = schema.paths
    assert '/user/repos' in paths

def test_get_operation(schema):
    op = schema.paths['/user/repos'].get
    assert op.responses

def test_get_operation_responses(schema):
    responses = schema.paths['/user/repos'].get.responses
    assert '200' in responses
    assert '403' in responses

def test_200_response(schema):
    response = schema.paths['/user/repos'].get.responses['200']
    assert 'x-dns-prefetch-control' in response.headers
    assert response.links == {}
    assert 'application/json' in response.content
    media_type = response.content['application/json']
    assert media_type.schema

def test_repos_content(schema):
    content = schema.paths['/user/repos'].get.responses['200'].content
    assert_that(content, has_key("application/json"))
    json_content = content['application/json']
    assert json_content.schema
    schema = json_content.schema
    assert schema._type == "array"
    # TODO Typeguard for Reference
    assert isinstance(schema.items, Schema)

def test_items(schema):
    schema = schema.paths['/user/repos'].get.responses['200'].content['application/json'].schema
    items = schema.items
    assert isinstance(items, Schema)  # typeguard
    assert isinstance(items.properties, dict)
    properties = items.properties
    assert_that(properties, has_entry('clone_url',
                                        equal_to(Schema(**{**_SCHEMA_DEFAULT, '_type': "string"}))))

def test_servers(schema):
    servers = schema.servers
    assert 1 == len(servers)
    assert 'http://api.github.com' == servers[0].url

def test_pokeapi_schema_valid(schema):
    # this should conflate to
    # /pokemon
    # /abilities/*
    # /types/*
    # /pokemon/*
    # meaning that it should recognize wildcards
    # from all these paths
    pokeapi_schema = build_schema_batch(pokeapi_requests, UpdateMode.GEN)
    paths = pokeapi_schema.paths.keys()
    assert 4 == len(paths)
    assert_that(paths, has_item("/v2/pokemon/"))
    assert_that(paths, has_item(
        matches_regexp(r'\/v2\/pokemon\/\{[\w]+\}')))
    assert_that(paths, has_item(
        matches_regexp(r'\/v2\/type\/\{[\w]+\}')))
    assert_that(paths, has_item(
        matches_regexp(r'\/v2\/ability\/\{[\w]+\}')))

#### petstore

get_pets_req = RequestBuilder.from_url(
    "http://petstore.swagger.io/v1/pets")
get_one_pet_req = RequestBuilder.from_url(
    "http://petstore.swagger.io/v1/pets/32")

pet_res = Response(bodyAsJson=None, timestamp=None, body="", statusCode=200, headers={})

@pytest.fixture
def get_one_pet_exchange():
    return HttpExchange(request=get_one_pet_req, response=pet_res)

@pytest.fixture
def get_pets_exchange():
    return HttpExchange(request=get_pets_req, response=pet_res)

@pytest.fixture
def orig_schema_paths():
    return list(PETSTORE_SCHEMA.paths.keys())

def test_update_respects_path_parameter(get_one_pet_exchange, orig_schema_paths):
    updated_schema = update_openapi(
        PETSTORE_SCHEMA, get_one_pet_exchange, UpdateMode.GEN)
    updated_schema_paths = list(updated_schema.paths.keys())
    assert_that(updated_schema_paths, is_(orig_schema_paths))

def test_update_respects_basepath(get_pets_exchange, orig_schema_paths):
    updated_schema = update_openapi(
        PETSTORE_SCHEMA, get_pets_exchange, UpdateMode.GEN)
    updated_schema_paths = list(updated_schema.paths.keys())
    assert_that(updated_schema_paths, is_(orig_schema_paths))


#####################
#### query params

query_req = RequestBuilder.from_url(
    url="https://petstore.swagger.io/v1/pets/32?id=1&car=ferrari")
query_res = Response(body="", statusCode=200, headers={}, bodyAsJson=None, timestamp=None)

@pytest.fixture
def query_exchange():
    return HttpExchange(request=query_req, response=query_res)

req_wo_query = RequestBuilder.from_url(
    url="https://petstore.swagger.io/v1/pets/32")
res_wo_query = Response(body="", statusCode=200, headers={}, bodyAsJson=None, timestamp=None)

@pytest.fixture
def exchange_wo_query():
    return HttpExchange(request=req_wo_query, response=res_wo_query)

@pytest.fixture
def expected_path_name():
    return "/v1/pets/32"

def test_build_with_query(query_exchange, expected_path_name):
    schema = build_schema_batch([query_exchange], UpdateMode.GEN)

    assert_that(list(schema.paths.keys()),
                is_([expected_path_name]))

    operation = schema.paths[expected_path_name].get

    assert operation.parameters

    parameters = operation.parameters

    assert_that(parameters, has_length(2))

    parameter_names = [param.name for param in parameters]

    assert_that(set(parameter_names), is_(set(['id', 'car'])))

def test_schema_update_with_query(exchange_wo_query, query_exchange, expected_path_name):
    schema = build_schema_batch([exchange_wo_query], UpdateMode.GEN)

    operation = schema.paths[expected_path_name].get
    assert operation.parameters == []

    updated_schema = update_openapi(schema, query_exchange, UpdateMode.GEN)

    operation = updated_schema.paths[expected_path_name].get
    assert_that(operation.parameters, has_length(2))

    first_query_param = operation.parameters[0]

    assert not first_query_param.required


text_request = RequestBuilder.from_url("https://example.com/v1")
text_response = Response(statusCode=200, body="Hello World", headers={}, bodyAsJson=None, timestamp=None)

@pytest.fixture
def text_exchange():
    return HttpExchange(request=text_request, response=text_response)

def test_build_string_body(text_exchange):
    schema = build_schema_batch([text_exchange], UpdateMode.GEN)
    response_content = schema.paths['/v1'].get.responses['200'].content

    assert_that(response_content, has_key("text/plain"))

    media_type = response_content["text/plain"]

    assert media_type.schema._type == "string"

def test_schema_in_replay_mode():
    reqs = []
    with open('tests/server/mock/callbacks/opbank/recordings/recording.jsonl','r') as rr: 
        reqs = rr.read().split('\n')

    reqs = [HttpExchangeBuilder.from_dict(json.loads(r)) for r in reqs if r != '']
    r = build_schema_batch(reqs, UpdateMode.REPLAY)
    # this schema has four recordings, of which two correspond to /v1/payments/confirm
    assert 2 == len(r.paths['/v1/payments/confirm'].post.responses['201'].content['application/json'].schema.oneOf)

import time
ACCEPTABLE_TIME = 10 # 10 seconds
def test_builder_speed():
    now = time.time()
    with open('resources/large.jsonl', 'r') as recordings:
        http_exchanges = [HttpExchangeBuilder.from_dict(json.loads(d)) for d in recordings.read().split('\n') if d != '']
        build_schema_online(iter(http_exchanges), mode=UpdateMode.REPLAY)
    assert (time.time() - now) < ACCEPTABLE_TIME